/* NinjaSCSI32Bi CardBus mode driver for Linux 2.4 */
/* YOKOTA Hiroshi <yokota@netlab.is.tsukuba.ac.jp> */

/* $Id: nsp32_cb.c,v 1.6 2002/02/08 06:47:43 elca Exp $ */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/string.h>
#include <linux/timer.h>
#include <linux/ioport.h>
#include <scsi/scsi.h>
#include <linux/major.h>
#include <linux/blk.h>
#include <linux/pci.h>
#include <linux/delay.h>

#include <linux/ctype.h>

#include <../drivers/scsi/scsi.h>
#include <../drivers/scsi/hosts.h>
#include <scsi/scsi_ioctl.h>


#include <pcmcia/driver_ops.h>

#include "nsp32_cb.h"


#define PCMCIA_DEBUG 9

#ifdef PCMCIA_DEBUG
static int pc_debug = PCMCIA_DEBUG;
MODULE_PARM(pc_debug, "i");
#define DEBUG(n, args...) if (pc_debug>(n)) printk(/*KERN_DEBUG*/ args)
static char *version =
"nsp32_cb (ver)";
#else
#define DEBUG(n, args...)
#endif

MODULE_LICENSE("GPL");

static int nsp32_detect(Scsi_Host_Template *sht);
static int nsp32_queuecommand(Scsi_Cmnd *SCpnt, void (*done)(Scsi_Cmnd *));
static const char *nsp32_info(struct Scsi_Host *shpnt);
static int nsp32_eh_bus_reset(Scsi_Cmnd *SCpnt);
static int nsp32_reset(Scsi_Cmnd *SCpnt, unsigned int reset_flags);
static int nsp32_release(struct Scsi_Host *shpnt);


typedef struct _nsp32_hw_data {
	int IrqNumber;
	int BaseAddress;
	int NumAddress;
#define NSP32_MMIO_OFFSET 0x0800
	unsigned long MmioAddress;


	Scsi_Cmnd *CurrentSC;

	struct pci_dev *Pci;

	char info_str[100];

//	char *autoparam;
} nsp32_hw_data;
static nsp32_hw_data nsp32_data;

#include "nsp32_io.h"


static Scsi_Host_Template driver_template = {
	proc_name:	       "nsp32_cb",
	name:			"WorkBit NinjaSCSI-32Bi",
	detect:                 nsp32_detect,
	info:                   nsp32_info,
	queuecommand:		nsp32_queuecommand,
	can_queue:		1,
	sg_tablesize:           SG_ALL,
        cmd_per_lun:            1,
	this_id:                7,
	use_clustering:         DISABLE_CLUSTERING,
	eh_bus_reset_handler:	nsp32_eh_bus_reset,
	reset:                  nsp32_reset,
	release:                nsp32_release,
};

#include "nsp32_debug.c"

static void nsp32_start_timer(Scsi_Cmnd *SCpnt, int time)
{
	unsigned int base = SCpnt->host->io_port;

	printk(__FUNCTION__ " time=%d\n", time);

	if (time & (~TIMER_CNT_MASK)) {
		printk("timer set overflow\n");
	}

	nsp32_write2(base, TIMER_SET, time & TIMER_CNT_MASK);
}

static unsigned int nsp32hw_start_selection(Scsi_Cmnd *SCpnt)
{
	unsigned int  host_id    = SCpnt->host->this_id;
	unsigned int  base       = SCpnt->host->io_port;
	unsigned char target     = SCpnt->target;
	unsigned char lun = SCpnt->lun;
	unsigned char phase, arbit;
//	nsp32_hw_data *data = &nsp32_data;
	int i;

	phase = nsp32_read1(base, SCSI_BUS_MONITOR);
	if (phase != BUSMON_BUS_FREE) {
		printk( " bus busy\n");
		show_busphase(phase & BUSMON_PHASE_MASK);
		return FALSE;
	}

	nsp32_write2(base, TRANSFER_CONTROL, ALL_COUNTER_CLR);
	nsp32_write2(base, SEL_TIME_OUT, 10000);
	nsp32_write1(base, SYNC_REG, 0);
	nsp32_write1(base, ACK_WIDTH, 0);

	nsp32_write2(base, COMMAND_CONTROL, CLEAR_CDB_FIFO_POINTER);
	for (i=0; i < SCpnt->cmd_len; i++) {
		nsp32_write1(base, COMMAND_DATA, SCpnt->cmnd[i]);
	}

	nsp32_write1(base, SCSI_OUT_LATCH_TARGET_ID, (1 << host_id) | (1 << target));

	nsp32_write4(base, SGT_ADR, 0);
	nsp32_write4(base, BM_START_ADR, 0);
	nsp32_write4(base, BM_CNT,       0);

	nsp32_write4(base, SCSI_MSG_OUT, (IDENTIFY(FALSE, lun) << 24) | MSGOUT_VALID | 1);

	nsp32_write2(base, TRANSFER_CONTROL, /*BM_START |*/ TRANSFER_GO | ALL_COUNTER_CLR | /*BM_SINGLE_MODE |*/ BLIEND_MODE | CB_IO_MODE);

	nsp32_write2(base, COMMAND_CONTROL, CLEAR_CDB_FIFO_POINTER | AUTOSCSI_START | AUTO_MSGIN_00_OR_04 | AUTO_MSGIN_02 | AUTO_ATN);

	i = 1000;
	do {
		arbit = nsp32_read1(base, ARBIT_STATUS);
		//printk(" arbit=0x%x\n", arbit);
		i--;
	} while ((arbit & (ARBIT_WIN | ARBIT_FAIL)) == 0 && i != 0);

	if((arbit & ARBIT_WIN) == 0) {
		printk( " arbit fail\n");
		nsp32_write1(base, SET_ARBIT, ARBIT_CLEAR);
		return FALSE;
        }

	nsp32_write1(base, SET_ARBIT, ARBIT_CLEAR);


	return TRUE;
}


static int nsp32_queuecommand(Scsi_Cmnd *SCpnt, void (*done)(Scsi_Cmnd *))
{
//	unsigned int base = SCpnt->host->io_port;
//	unsigned char target = SCpnt->target;
	nsp32_hw_data *data = &nsp32_data;
//	int i;

//	DEBUG(0, __FUNCTION__ "() SCpnt=0x%p target=%d lun=%d buff=0x%p bufflen=%d use_sg=%d\n",
//	      SCpnt, target, SCpnt->lun, SCpnt->request_buffer, SCpnt->request_bufflen, SCpnt->use_sg);

	if (data->CurrentSC != NULL ) {
		printk("Currentsc != NULL\n");
		SCpnt->result   = DID_NO_CONNECT << 16;
		done(SCpnt);

		return -1;
	}


//	show_command(SCpnt);

	SCpnt->scsi_done     = done;
	data->CurrentSC      = SCpnt;
	SCpnt->SCp.Status    = CHECK_CONDITION;
	SCpnt->SCp.Message   = 0;
	SCpnt->resid         = SCpnt->request_bufflen;

	/* setup scratch area
	   SCp.ptr		: buffer pointer
	   SCp.this_residual	: buffer length
	   SCp.buffer		: next buffer
	   SCp.buffers_residual : left buffers in list
	   SCp.phase		: current state of the command */
	if (SCpnt->use_sg) {
		SCpnt->SCp.buffer	    = (struct scatterlist *) SCpnt->request_buffer;
		SCpnt->SCp.ptr		    = SCpnt->SCp.buffer->address;
		SCpnt->SCp.this_residual    = SCpnt->SCp.buffer->length;
		SCpnt->SCp.buffers_residual = SCpnt->use_sg - 1;
	} else {
		SCpnt->SCp.ptr		    = (char *) SCpnt->request_buffer;
		SCpnt->SCp.this_residual    = SCpnt->request_bufflen;
		SCpnt->SCp.buffer	    = NULL;
		SCpnt->SCp.buffers_residual = 0;
	}


//	nsp32_eh_bus_reset(SCpnt); // xxx

	if (nsp32hw_start_selection(SCpnt) != TRUE) {
		SCpnt->result   = DID_NO_CONNECT << 16;
		data->CurrentSC = NULL;
		SCpnt->scsi_done(SCpnt);

		return -1;
	}



	return 0;
}

static int nsp32hw_init(struct Scsi_Host *host)
{
	unsigned int  base = host->io_port;
	unsigned short irq_stat;
	unsigned char power;

	printk("index 0x%x\n", nsp32_read2(base, INDEX_REG));

	nsp32_write2(base, IRQ_CONTROL,      IRQ_CONTROL_ALL_IRQ_MASK);
	nsp32_write2(base, TRANSFER_CONTROL, 0);
	nsp32_write4(base, BM_CNT,           0);

	do {
		irq_stat = nsp32_read2(base, IRQ_STATUS);
	} while (irq_stat & 0x0ff0);
//	printk("irq_stat 0x%x\n", irq_stat);

	nsp32_index_write1(base, FIFO_FULL_SHLD_COUNT, 0x40);
	nsp32_index_write1(base, FIFO_EMPTY_SHLD_COUNT, 0x40);

	//printk("full 0x%x emp 0x%x\n", nsp32_index_read1(base, FIFO_FULL_SHLD_COUNT), nsp32_index_read1(base, FIFO_EMPTY_SHLD_COUNT));

//	nsp32_index_write1(base, CLOCK_DIV, 2);

	nsp32_write1(base, PARITY_CONTROL, 0);
	nsp32_index_write1(base, MISC_WR, SCSI_DIRECTION_DETECTOR_SELECT |
			   MASTER_TERMINATION_SELECT |
			   AUTOSEL_TIMING_SEL );
//	nsp32_index_write1(base, BM_CYCLE, MEMRD_CMD1 | SGT_AUTO_PARA_MEMED_CMD);

	nsp32_index_write1(base, TERM_PWR_CONTROL, 0);
	power = nsp32_index_read1(base, TERM_PWR_CONTROL);
	if (!(power & SENSE)) {
		printk("term power on\n");
		nsp32_index_write1(base, TERM_PWR_CONTROL, BPWR);
	}

	nsp32_write2(base, TIMER_SET, TIMER_STOP);
	nsp32_write2(base, TIMER_SET, TIMER_STOP);

	nsp32_write1(base, SYNC_REG, 0);
	nsp32_write1(base, ACK_WIDTH, 0);


	nsp32_index_write2(base, IRQ_SELECT, IRQSELECT_TIMER_IRQ |
			                     IRQSELECT_SCSIRESET_IRQ |
			                     IRQSELECT_FIFO_SHLD_IRQ |
			                     IRQSELECT_RESELECT_IRQ |
			                     IRQSELECT_PHASE_CHANGE_IRQ |
			                     IRQSELECT_AUTO_SCSI_SEQ_IRQ |
			   //IRQSELECT_BMCNTERR_IRQ |
			                     IRQSELECT_TARGET_ABORT_IRQ |
			   IRQSELECT_MASTER_ABORT_IRQ );
	nsp32_write2(base, IRQ_CONTROL, 0);

	return TRUE;
}

static void nsp32_pio_read(Scsi_Cmnd *SCpnt)
{
	unsigned int base = SCpnt->host->io_port;
	int timeout;
	int cnt;
	unsigned long tmp;
	unsigned short i, fifo_stat;
	unsigned char busphase;

	//printk(__FUNCTION__ "\n");

	timeout = 10;

	fifo_stat = nsp32_read2(base, IRQ_STATUS);
	//nsp32_write2(base, IRQ_CONTROL, (fifo_stat & IRQ_CONTROL_ALL_IRQ_MASK) | FIFO_IRQ_MASK);

	while ((SCpnt->SCp.this_residual > 0 ||
		SCpnt->SCp.buffers_residual > 0 ) &&
	       (timeout-- > 0)) {

		if (SCpnt->resid == 0) {
			printk("resid=0\n");
			break;
		}

		i = nsp32_read2(base, FIFO_REST_CNT);
		busphase = nsp32_read1(base, SCSI_BUS_MONITOR) & BUSMON_PHASE_MASK;

		if (((i & FIFO_REST_MASK) == 0) &&
		    ((i & FIFO_FULL_SHLD_FLAG) == 0)) {
			if (busphase == BUSPHASE_DATA_IN) {
				//printk("c");
				continue;
			} else {
				//printk("phase change\n");
				//show_busphase(busphase);
				break;
			}
		}

		//printk("i 0x%x this 0x%x\n", i, SCpnt->SCp.this_residual);

		if(SCpnt->SCp.this_residual > 3) {
			cnt = MIN(i & FIFO_REST_MASK, SCpnt->SCp.this_residual) & ~3;

			//printk("read 0x%x\n", cnt);

			nsp32_fifo_read(base, SCpnt->SCp.ptr, cnt >> 2);

		} else {
			tmp = le32_to_cpu(nsp32_read4(base, FIFO_DATA_LOW));

			printk("r %d tmp=0x%08lx\n",SCpnt->SCp.this_residual, tmp);

			switch(SCpnt->SCp.this_residual) {
			case 3:
				//printk("3\n");
				cnt = 3;
				SCpnt->SCp.ptr[0] = (tmp & 0x000000ff) >> 0;
				SCpnt->SCp.ptr[1] = (tmp & 0x0000ff00) >> 8;
				SCpnt->SCp.ptr[2] = (tmp & 0x00ff0000) >> 16;

				break;
			case 2:
				//printk("2\n");
				cnt = 2;
				SCpnt->SCp.ptr[0] = (tmp & 0x000000ff) >> 0;
				SCpnt->SCp.ptr[1] = (tmp & 0x0000ff00) >> 8;
				break;
			case 1:
				//printk("1\n");
				cnt = 1;
				SCpnt->SCp.ptr[0] = (tmp & 0x000000ff) >> 0;
				break;
			default:
				cnt = 0;
				printk("this = 0???\n");
				break;
			}
		}

		SCpnt->SCp.ptr += cnt;
		SCpnt->SCp.this_residual -= cnt;
		SCpnt->resid -= cnt;

		if (SCpnt->SCp.this_residual	== 0 &&
		    SCpnt->SCp.buffers_residual != 0 ) {
			SCpnt->SCp.buffers_residual--;
			SCpnt->SCp.buffer++;
			SCpnt->SCp.ptr		 = SCpnt->SCp.buffer->address;
			SCpnt->SCp.this_residual = SCpnt->SCp.buffer->length;

			//printk("next\n");
			timeout = 10;
		}

	}

	//nsp32_write2(base, IRQ_CONTROL, (fifo_stat & IRQ_CONTROL_ALL_IRQ_MASK) & ~FIFO_IRQ_MASK);

	//printk(__FUNCTION__ " out timeout %d\n", timeout);
}

static void nsp32_pio_write(Scsi_Cmnd *SCpnt)
{
	unsigned int base = SCpnt->host->io_port;
	int timeout;
	int cnt;
	unsigned long tmp;
	unsigned short i, fifo_stat;
	unsigned char busphase;

	timeout = 10;

	fifo_stat = nsp32_read2(base, IRQ_STATUS);
	//nsp32_write2(base, IRQ_CONTROL, (fifo_stat & IRQ_CONTROL_ALL_IRQ_MASK) | FIFO_IRQ_MASK);

	while (((SCpnt->SCp.this_residual > 0) || (SCpnt->SCp.buffers_residual > 0) ) &&
	       (timeout-- > 0)) {

		if (SCpnt->resid == 0) {
			break;
		}

		i = nsp32_read2(base, FIFO_REST_CNT);
		busphase = nsp32_read1(base, SCSI_BUS_MONITOR) & BUSMON_PHASE_MASK;

		if (((i & FIFO_REST_CNT) != 0) &&
		    ((i & FIFO_EMPTY_SHLD_FLAG) == 0)) {
			if (busphase == BUSPHASE_DATA_OUT) {
				//printk("c");
				continue;
			} else {
				//printk("phase change\n");
				//show_busphase(busphase);
				break;
			}
		}

		if(SCpnt->SCp.this_residual > 3) {
			cnt = MIN(SCpnt->SCp.this_residual, (FIFO_MAX / 2)) & ~3;

			//printk("write 0x%x\n", cnt);

			nsp32_fifo_write(base, SCpnt->SCp.ptr, cnt >> 2);

		} else {
			switch(SCpnt->SCp.this_residual) {
			case 3:
				//printk("3\n");
				cnt = 3;
				tmp = SCpnt->SCp.ptr[0] << 0 |
					SCpnt->SCp.ptr[1] << 8 |
					SCpnt->SCp.ptr[2] << 16;
				break;
			case 2:
				//printk("2\n");
				cnt = 2;
				tmp = SCpnt->SCp.ptr[0] << 0 |
					SCpnt->SCp.ptr[1] << 8;
				break;
			case 1:
				//printk("1\n");
				cnt = 1;
				tmp = SCpnt->SCp.ptr[0] << 0;
				break;
			default:
				cnt = 0;
				printk("this = 0???\n");
				tmp = 0;
				break;
			}

			printk("w %d tmp=0x%lx\n",SCpnt->SCp.this_residual, tmp);

			nsp32_write4(base, FIFO_DATA_LOW, cpu_to_le32(tmp));

		}

		SCpnt->SCp.ptr += cnt;
		SCpnt->SCp.this_residual -= cnt;
		SCpnt->resid -= cnt;

		if (SCpnt->SCp.this_residual	== 0 &&
		    SCpnt->SCp.buffers_residual != 0 ) {
			SCpnt->SCp.buffers_residual--;
			SCpnt->SCp.buffer++;
			SCpnt->SCp.ptr		 = SCpnt->SCp.buffer->address;
			SCpnt->SCp.this_residual = SCpnt->SCp.buffer->length;
			timeout = 10;
		}
	}

	//nsp32_write2(base, IRQ_CONTROL, (fifo_stat & IRQ_CONTROL_ALL_IRQ_MASK) & ~FIFO_IRQ_MASK);
}


static void nsp32intr(int irq, void *dev_id, struct pt_regs *regs)
{
	nsp32_hw_data *data = dev_id;
	unsigned int base;
	Scsi_Cmnd *SCpnt;
	unsigned short auto_stat, irq_stat, trans_stat;
	unsigned char busmon, busphase;

	base = data->BaseAddress;

	irq_stat = nsp32_read2(base, IRQ_STATUS);
	if ((irq_stat & IRQSTATUS_ANY_IRQ) == 0) {
		printk(__FUNCTION__ " irq other\n");
		return;
	}
	trans_stat = nsp32_read2(base, TRANSFER_STATUS);
	if ((irq_stat == 0xffff) && (trans_stat == 0xffff)) {
		printk(__FUNCTION__ " card disconnect\n");
		if (data->CurrentSC != NULL) {
			printk(" clean up current SCSI command\n");
			SCpnt           = data->CurrentSC;
			data->CurrentSC = NULL;
			SCpnt->result   = DID_BAD_TARGET << 16;
			SCpnt->scsi_done(SCpnt);
		}
		return;
	}

	if (irq_stat & IRQSTATUS_TIMER_IRQ) {
		printk("timer stop\n");
		nsp32_write2(base, TIMER_SET, TIMER_STOP);
		return;
	}

	//printk("busmon 0x%x\n",busmon);

	SCpnt = data->CurrentSC;

//	printk(__FUNCTION__ " dev_id=0x%p irq_stat=0x%x trans_stat=0x%x\n", dev_id, irq_stat, trans_stat);

//	printk("fifo rest 0x%x\n", nsp32_read2(base, FIFO_REST_CNT));

//	if (irq_stat & IRQSTATUS_FIFO_SHLD_IRQ) {
//		printk("fifo\n");
//		return;
//	}




//	nsp32_write2(base, TRANSFER_CONTROL, trans_stat & ~BM_START);

	if(SCpnt == NULL) {
		printk("SCpnt==NULL\n");
		printk("irq_stat=0x%x trans_stat=0x%x\n", irq_stat, trans_stat);
		return;
	}

	if(irq_stat & IRQSTATUS_AUTOSCSI_IRQ){
		auto_stat = nsp32_read2(base, SCSI_EXECUTE_PHASE);


		if(auto_stat & SELECTION_TIMEOUT) {
//			printk("sel timeout\n");
			data->CurrentSC = NULL;
			SCpnt->result   = DID_NO_CONNECT << 16;
			SCpnt->scsi_done(SCpnt);
			return;
		}




		if ((auto_stat & DATA_IN_PHASE) &&
		    (SCpnt->resid >  0) &&
		    (nsp32_read2(base, FIFO_REST_CNT) & FIFO_REST_MASK) != 0) {
			//printk("auto+fifo\n");
			nsp32_pio_read(SCpnt);
		}

		if(auto_stat & MSGIN_04_VALID){
			SCpnt->SCp.Status = nsp32_read1(base, SCSI_CSB_IN);
			SCpnt->SCp.Message = 4;
			printk("disconnect\n");
			return;
		}
		if(auto_stat & MSGIN_00_VALID){
			SCpnt->SCp.Status = nsp32_read1(base, SCSI_CSB_IN);
			SCpnt->SCp.Message = 0;

			//printk("normal end stat=0x%x resid=0x%x\n", SCpnt->SCp.Status, SCpnt->resid);

			data->CurrentSC = NULL;
			SCpnt->result   = (DID_OK << 16) | (SCpnt->SCp.Message << 8) | (SCpnt->SCp.Status << 0);
			SCpnt->scsi_done(SCpnt);
			return;
		}

		if(auto_stat & MSG_OUT_OCCUER) {
			printk("msgout occuer\n");

			nsp32_write4(base, SCSI_MSG_OUT, (IDENTIFY(FALSE, SCpnt->lun) << 24) | MSGOUT_VALID | 1);

			//nsp32_write2(base, COMMAND_CONTROL, CLEAR_CDB_FIFO_POINTER |
			//	     AUTOSCSI_RESTART       |
			//	     AUTO_MSGIN_00_OR_04    |
			//	     AUTO_MSGIN_02          |
			//	     AUTO_ATN);
			return;
		}
		if(auto_stat & MSG_IN_OCCUER) {
			printk("msgin occuer\n");

			SCpnt->SCp.Message = nsp32_read1(base, SCSI_DATA_WITH_ACK);

			//nsp32_write2(base, COMMAND_CONTROL, CLEAR_CDB_FIFO_POINTER |
			//	     AUTOSCSI_RESTART       |
			//	     AUTO_MSGIN_00_OR_04    |
			//	     AUTO_MSGIN_02          |
			//	     AUTO_ATN);
			return;
		}




		printk("autoscsi other\n");
		show_autophase(auto_stat);

		data->CurrentSC = NULL;
		SCpnt->result = DID_BAD_TARGET << 16;
		SCpnt->scsi_done(SCpnt);

		return;
	}


	busmon = nsp32_read1(base, SCSI_BUS_MONITOR);
	busphase = busmon & BUSMON_PHASE_MASK;
//	show_busphase(busphase);


	if ((irq_stat & IRQSTATUS_FIFO_SHLD_IRQ) ||
	    (irq_stat & IRQSTATUS_PHASE_CHANGE_IRQ)) {
		//printk("fifo\n");
		switch(busphase) {
		case BUSPHASE_DATA_OUT:

			nsp32_pio_write(SCpnt);

			break;

		case BUSPHASE_DATA_IN:

			nsp32_pio_read(SCpnt);

			break;
		default:
			printk("fifo/other phase?\n");
			show_busphase(busphase);
			break;
		}

		//nsp32_write2(base, COMMAND_CONTROL, CLEAR_CDB_FIFO_POINTER |
		//	     AUTOSCSI_RESTART       |
		//	     AUTO_MSGIN_00_OR_04    |
		//	     AUTO_MSGIN_02          |
		//	     AUTO_ATN);
		return;
	}

	printk("irq_stat=0x%x trans_stat=0x%x\n", irq_stat, trans_stat);

	switch (busphase) {
	case BUSPHASE_MESSAGE_OUT:
		printk("msgout\n");

		break;
	case BUSPHASE_MESSAGE_IN:
		printk("msgin\n");

		break;
	case BUSPHASE_STATUS:
		printk("status\n");

		break;
	default:
		printk("?/other phase?\n");
		show_busphase(busphase);
		break;
	}

	//printk("restart\n");

	nsp32_write2(base, COMMAND_CONTROL, CLEAR_CDB_FIFO_POINTER |
		     AUTOSCSI_RESTART       |
		     AUTO_MSGIN_00_OR_04    |
		     AUTO_MSGIN_02          |
		     AUTO_ATN);
	return;


}

static int nsp32_detect(Scsi_Host_Template *sht)
{
	struct Scsi_Host *host;	/* registered host structure */
	int ret;
	nsp32_hw_data *data = &nsp32_data;

//	printk(__FUNCTION__ "\n");

	host		  = scsi_register(sht, 0);
	host->irq         = data->IrqNumber;
	host->io_port	  = data->BaseAddress;
	host->unique_id	  = data->BaseAddress;
	host->n_io_port	  = data->NumAddress;
	host->base        = data->MmioAddress;

	scsi_set_pci_device(host, data->Pci);


//	data->autoparam   = kmalloc(4*15, GFP_ATOMIC);


	nsp32hw_init(host);

	snprintf(data->info_str, sizeof(data->info_str), "nsp32: irq %d, io 0x%lx+0x%x", host->irq, host->io_port, host->n_io_port);

	sht->name	  = data->info_str;

	ret = request_irq(host->irq, nsp32intr, SA_SHIRQ, "nsp32_cb", data);
	if (ret < 0) {
		printk("req irq fail\n");
		return 0;
	}

	return 1;
}

static int nsp32_release(struct Scsi_Host *shpnt)
{
	nsp32_hw_data *data = &nsp32_data;

//	kfree(data->autoparam);

	printk("free irq\n");
	if (shpnt->irq)
		free_irq(shpnt->irq, data);

	printk("free io\n");
	if (shpnt->io_port && shpnt->n_io_port)
		release_region(shpnt->io_port, shpnt->n_io_port);
	return 0;
}

static const char *nsp32_info(struct Scsi_Host *shpnt)
{
	nsp32_hw_data *data = &nsp32_data;

//	printk(__FUNCTION__ "\n");

	return data->info_str;
}


static int nsp32_reset(Scsi_Cmnd *SCpnt, unsigned int reset_flags)
{
        printk(__FUNCTION__ " SCpnt=0x%p why=%d\n", SCpnt, reset_flags);

        nsp32_eh_bus_reset(SCpnt);

        return SCSI_RESET_SUCCESS | SCSI_RESET_BUS_RESET;
}


static int nsp32_eh_bus_reset(Scsi_Cmnd *SCpnt)
{
	unsigned int base = SCpnt->host->io_port;
	int	     i;

	printk(__FUNCTION__ "()\n");

	nsp32_write2(base, IRQ_CONTROL, IRQ_CONTROL_ALL_IRQ_MASK);


	nsp32_write2(base, TRANSFER_CONTROL, 0);
	nsp32_write4(base, BM_CNT,           0);


	nsp32_write1(base, SCSI_BUS_CONTROL, BUSCTL_RST);
	mdelay(100);
	nsp32_write1(base, SCSI_BUS_CONTROL, 0);
	for(i = 0; i < 5; i++) {
                nsp32_read2(base, IRQ_STATUS); /* dummy read */
        }
	nsp32_write2(base, IRQ_CONTROL, 0);

	return SUCCESS;
}



static int __devinit nsp32_probe(struct pci_dev *pdev, const struct pci_device_id *id)
{
	int ret;
	nsp32_hw_data *data = &nsp32_data;


	printk(__FUNCTION__ "in\n");


        ret = pci_enable_device(pdev);
	if (ret) {
		printk("ret\n");
		return ret;
	}

	data->IrqNumber = pdev->irq;
	data->BaseAddress = pci_resource_start(pdev, 0);
	data->NumAddress = pci_resource_len(pdev, 0);
	data->MmioAddress = (unsigned long)ioremap_nocache(pci_resource_start(pdev, 1), pci_resource_len(pdev, 1));
	data->Pci = pdev;

	scsi_register_module(MODULE_SCSI_HA, &driver_template);


	printk(KERN_INFO "nsp32 cardbus at irq %i mmio 0x%lx\n", pdev->irq, data->MmioAddress);

	printk("end\n");
	return 0;
}

static void __devexit nsp32_remove(struct pci_dev *pdev)
{
	nsp32_hw_data *data = &nsp32_data;

	printk(__FUNCTION__ " in\n");
	

	scsi_unregister_module(MODULE_SCSI_HA, &driver_template);

	if (data->MmioAddress != 0)
		iounmap((void *)(data->MmioAddress));
}




static struct pci_device_id nsp32_pci_table[] __devinitdata = {
	{
		vendor: 0x10fc,
		device: 0x0005,
		subvendor: PCI_ANY_ID,
		subdevice: PCI_ANY_ID,
		//class: 0,
		//class_mask: 0,
		//driver_data: 0,
	},
	{
		vendor: 0x1145,
		device: 0xf007,
		subvendor: PCI_ANY_ID,
		subdevice: PCI_ANY_ID,
		//class: 0,
		//class_mask: 0,
		//driver_data: 0,
	},
	{
		vendor: 0x1145,
		device: 0x8007,
		subvendor: PCI_ANY_ID,
		subdevice: PCI_ANY_ID,
		//class: 0,
		//class_mask: 0,
		//driver_data: 0,
	},
	{0,0,},
};
MODULE_DEVICE_TABLE(pci, nsp32_pci_table);

static struct pci_driver nsp32_driver = {
	name: "nsp32_cb",
	id_table: nsp32_pci_table,
	probe: nsp32_probe,
	remove: nsp32_remove,
	save_state: NULL,
	suspend: NULL,
	resume: NULL,
	enable_wake: NULL,
};




static int __init init_nsp32_cb(void) {
	DEBUG(0, "%s\n", version);
	return pci_module_init(&nsp32_driver);
}

static void __exit exit_nsp32_cb(void) {
	DEBUG(0, "nsp32_cs: unloading\n");
	pci_unregister_driver(&nsp32_driver);
	printk(__FUNCTION__ " out\n");
}

module_init(init_nsp32_cb);
module_exit(exit_nsp32_cb);

